/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: carol@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: MasterServerClusterViewManager.java 1892 2008-07-07 12:54:22Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.impl.coordinator.master;

import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicSession;
import javax.jms.TopicSubscriber;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import net.jcip.annotations.ThreadSafe;

import org.objectweb.joram.client.jms.Topic;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.objectweb.joram.client.jms.admin.User;
import org.ow2.cmi.controller.server.DistributedObjectInfo;
import org.ow2.cmi.controller.server.impl.coordinator.CoordinatedServerClusterViewManager;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.cmi.reference.ServerId;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * @author Loris Bouzonnet
 */
@ThreadSafe
public class MasterServerClusterViewManager extends CoordinatedServerClusterViewManager implements MessageListener {

    /**
     * Logger.
     */
    private static final Log logger = LogFactory
            .getLog(MasterServerClusterViewManager.class);

    private TopicSubscriber subscriber;

    private int delayToRefresh;

    private final ConcurrentHashMap<String, List<CMIReference>> objRefs =
        new ConcurrentHashMap<String, List<CMIReference>>();

    private final ConcurrentHashMap<String, DistributedObjectInfo> objInfos =
        new ConcurrentHashMap<String, DistributedObjectInfo>();

    @Override
    protected void initServerConfig(final String contextFactory, final ServerId serverId, final Context cmiContext) {
        TopicConnection tc = getTopicConnection();
        try {
            if(tc == null) {
                Hashtable<String, String> env = new Hashtable<String, String>();
                env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
                env.put(Context.PROVIDER_URL, serverId.getProviderURL());
                Context context = new InitialContext(env);
                TopicConnectionFactory tcf = (TopicConnectionFactory) context.lookup("JTCF");
                tc = tcf.createTopicConnection("master", "cmi");
                setTopicConnection(tc);
                TopicSession topicSession = tc.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
                setSession(topicSession);
                subscriber = topicSession.createSubscriber(getTopic());
                subscriber.setMessageListener(this);

                context.rebind("cmiCoordinator", getTopic());
                context.close();
                tc.start();
            }
        } catch (NamingException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (JMSException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        super.initServerConfig(contextFactory, serverId, cmiContext);
    }

    @Override
    protected boolean containObject(final String objectName) {
        return objRefs.containsKey(objectName);
    }

    @Override
    protected DistributedObjectInfo getDistributedObjectInfo(final String objectName)
            throws ObjectNotFoundException {
        return objInfos.get(objectName);
    }

    @Override
    protected void addDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        objInfos.putIfAbsent(objectName, distributedObjectInfo);
    }

    /**
     * Sets informations on the clustered object with the given name.
     *
     * @param objectName
     *            a name of object
     * @param distributedObjectInfo
     *            informations on the clustered object
     */
    @Override
    protected void setDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        objInfos.put(objectName, distributedObjectInfo);
    }

    @Override
    public void addCMIReference(final CMIReference cmiReference) {
        List<CMIReference> emptyRefs = Collections.synchronizedList(new ArrayList<CMIReference>());
        List<CMIReference> refs = objRefs.putIfAbsent(cmiReference.getObjectName(), emptyRefs);
        if(refs != null) {
            refs.add(cmiReference);
        } else {
            emptyRefs.add(cmiReference);
        }
    }


    @Override
    public void doStart() {
        // TODO Allow to create a clustered topic
        try {
            AdminModule.connect("root", "root", 60);
            Topic topic = Topic.create("mbeancmd_events");
            User master = User.create("master", "cmi");
            topic.setReader(master);
            User slave = User.create("slave", "cmi");
            topic.setWriter(slave);
            setTopic(topic);
            AdminModule.disconnect();
        } catch (ConnectException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (UnknownHostException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (AdminException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }


    @Override
    public void doStop() {
        // TODO Auto-generated method stub

    }

    @Override
    public List<CMIReference> getCMIReferences(final String objectName)
            throws ObjectNotFoundException {
        return objRefs.get(objectName);
    }

    @Override
    public List<CMIReference> getCMIReferences(final String objectName,
            final String protocolName, final Set<Object> qualities)
            throws ObjectNotFoundException {
        return getCMIReferences(objectName);
    }


    @Override
    public Set<String> getObjectNames() {
        return objRefs.keySet();
    }

    @Override
    public void removeCMIReference(final CMIReference cmiReference) {
        objRefs.remove(cmiReference.getObjectName(), cmiReference);
    }

    @Override
    public void setDelayToRefresh(final int delay) {
        delayToRefresh = delay;
    }

    @Override
    public int getDelayToRefresh() {
        return delayToRefresh;
    }

    public void onMessage(final Message msg) {
        try {
            if(msg instanceof ObjectMessage) {
                ObjectMessage objectMessage = (ObjectMessage) msg;
                Object object = objectMessage.getObject();
                if(object instanceof CMIReference) {
                    CMIReference cmiRef = (CMIReference) object;
                    if(objectMessage.getBooleanProperty("add")) {
                        logger.debug("A new object is available at this location: {0}", cmiRef);
                        addCMIReference(cmiRef);
                    } else {
                        logger.debug("A new object is to remove for this location: {0}", cmiRef);
                        removeCMIReference(cmiRef);
                    }
                } else if(object instanceof DistributedObjectInfo) {
                    DistributedObjectInfo distributedObjectInfo = (DistributedObjectInfo) object;
                    if(objectMessage.getBooleanProperty("add")) {
                        addDistributedObjectInfo(distributedObjectInfo.getObjectName(), distributedObjectInfo);
                    } else {
                        setDistributedObjectInfo(distributedObjectInfo.getObjectName(), distributedObjectInfo);
                    }
                }
            }
        } catch (JMSException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

}
